@@ -24,7 +24,10 @@ import time
 import json
 import queue
 import threading
+import gc # Garbage Collector
+import socket
 
+__version__ = '0.6.0'
 
 module_logger = logging.getLogger(__name__)
 
@@ -44,6 +47,7 @@ class GAAPItoDB(object):
                         apiQuota=0,
                         metrics=None,
                         dateRangePartitionSize=None,  # in days
+                        dbWritePartitions=None,
                         emptyRows=True,
                         dbURL=None,
                         targetTable=None,
@@ -77,13 +81,12 @@ class GAAPItoDB(object):
         self.dbURL=dbURL
         self.targetTable=targetTable
         self.dateRangePartitionSize=dateRangePartitionSize
+        self.dbWritePartitions=dbWritePartitions
         self.quotaControl={}
         self.restart=restart
         self.update=update
         self.apiQuota=apiQuota
 
-        self.start=start
-        self.end=end
         self.incremental=incremental
         self.endLag=endLag
         
@@ -103,6 +106,13 @@ class GAAPItoDB(object):
             # Lets discover timezone configured for this GA View
             self.gaTimezone=self.gaViewObject['timezone']
 
+        self.start = pd.Timestamp(start,tz=self.gaTimezone).to_pydatetime()
+        
+        self.end = end
+        if end:
+            self.end=pd.Timestamp(end,tz=self.gaTimezone).to_pydatetime()
+
+
         # Show some info about timezone, causae its an important subject
         t=pd.Timestamp.now()
         self.logger.warning("This GA View's time zone is «{tz}». A timestamp as «{example}» will be handled as «{example_converted}» and stored in database in UTC format, as «{utc}».".format(
@@ -374,8 +384,8 @@ class GAAPItoDB(object):
         for p in periods:
             ranges.append(
                 [
-                    p.start_time.to_pydatetime(),
-                    p.end_time.to_pydatetime()
+                    p.start_time.tz_localize(self.gaTimezone).to_pydatetime(warn=False),
+                    p.end_time.tz_localize(self.gaTimezone).to_pydatetime(warn=False)
                 ]
             )
         
@@ -412,20 +422,17 @@ class GAAPItoDB(object):
         
         # Some debug messages
         
-        self.logger.debug("GA call count since {}: {}".format(self.quotaControl['lastStart'], self.quotaControl['count']))
         self.logger.debug("Query for GA: {}".format(json.dumps(body)))
-        
-        
-        
-        # Finaly call Google Analytics API
-        
-        try:
-            report = self.ga.reports().batchGet(body=body, quotaUser=self.processor).execute()
-        except Exception as e:
-            self.logger.error("Unrecoverable error while calling GA: " + e)
-
-        
-        self.quotaControl['count']+=1
+                
+        # Finaly call Google Analytics API; retry ad infinitum if we get timeout exceptions.
+        while True:
+            try:
+                self.logger.debug("GA call count since {}: {}".format(self.quotaControl['lastStart'], self.quotaControl['count']))
+                self.quotaControl['count'] += 1
+                report = self.ga.reports().batchGet(body=body, quotaUser=self.processor).execute()
+                break
+            except Exception as e:
+                self.logger.warning("GA timed out, broken pipe or other error; retrying…")
 
         return report
 
@@ -453,6 +460,7 @@ class GAAPItoDB(object):
         
         subreports = self.subreportDimensions()
         timepartitions = self.getDateRangePartitions()
+        keys = self.getReportKeys()
 
         self.logger.debug(f'Subreport indexes: {subreports}')
         self.logger.debug(f'Time partitions to cover entire period requested: {timepartitions}')
@@ -515,12 +523,32 @@ class GAAPItoDB(object):
 
                 while cont:
                     # Iterate over subreport pages of about 100000 rows
+
+                    s = max(p[0],self.effectiveStart)
+                    e = min(p[1],self.end)
+
+                    self.logger.debug('Working on subreport for {focus}+{keys} ({pos} of {tot}), page {page}, time range of {start} ➔ {end}'.format(
+                            focus=list(set(dimTitles).difference(keys)),
+                            keys=keys,
+                            pos=i+1,
+                            tot=len(subreports),
+                            page=pageiteration,
+                            start=s,
+                            end=e
+                        )
+                    )
                     
-                    self.logger.debug("Working on subreport {} of {}: 【{}】 【{}】 【{}】…".format(i+1,len(subreports),timePartitionName,dimTitles,pageiteration))
+#                     self.logger.debug("Working on subreport {} of {}: 【{}】 【{}】 【{}】…".format(i+1,len(subreports),timePartitionName,dimTitles,pageiteration))
                     
                     if pageiteration>0:
                         query['pageToken'] = nextPageToken
 
+                    try:
+                        # Free big objects in RAM
+                        del report
+                    except NameError:
+                        pass
+                    
                     report = self.callGA(
                         body={
                             'reportRequests': [query],
@@ -556,7 +584,7 @@ class GAAPItoDB(object):
                         self.logger.debug("Subreport page size has {} rows.".format(len(report['reports'][0]['data']['rows'])))
 
                         if samplesReadCount:
-                            self.logger.debug("Sample space size: {}. Samples read: {}. Read {}% of sample space.".format(samplingSpaceSize,samplesReadCount,100*samplesReadCount/samplingSpaceSize))
+                            self.logger.warning("Sample space size: {}. Samples read: {}. Read {}% of sample space.".format(samplingSpaceSize,samplesReadCount,100*samplesReadCount/samplingSpaceSize))
                         else:
                             self.logger.debug("Data is complete and not sampled !")
 
@@ -574,15 +602,22 @@ class GAAPItoDB(object):
                     columns=self.dimensionItemsToList('title', filter=subreports[i]),
                     data=result
                 ))
-                self.logger.debug("Subreport shape size is {}×{}".format(self.subreports[-1].shape[0],self.subreports[-1].shape[1]))
+                self.logger.debug("Subreport shape size is {}×{}".format(
+                    self.subreports[-1].shape[0],
+                    self.subreports[-1].shape[1])
+                )
 
                 # Calculate a wanna-be-unique hash for each line based on key columns/dimensions
-                keys=self.getReportKeys()
                 self.subreports[-1] = GAAPItoDB.makePrimaryKey(self.subreports[-1],keys)
 
+
+#                 buffer = io.StringIO()
+#                 self.subreports[-1].info(verbose=True, buf=buffer)
+#                 self.logger.debug("Subreport memory profile:\n{}".format(buffer.getvalue()))
                 
                 # Free some RAM
                 del result
+                
 
                 # At this point all pages of a subreport inside a time partition were read and stored in a subreport DataFrame.
                 # Continue to next subreport for same time partition.
@@ -607,17 +642,38 @@ class GAAPItoDB(object):
                                 sort=False
                     )
 
-                # Coalesce values of key columns
-                for k in keys:
-                    for i in range(1, len(self.subreports)):
+                    # Coalesce values of key columns so the non-“__{i}” ones will have the data
+                    for k in keys:
                         self.report[k]=self.report[k].combine_first(self.report[f'{k}__{i}'])
+                    
+                    # Delete overlapping columns
+                    cols=self.report.columns
+                    todrop=[]
+                    for c in cols:
+                        if f"__{i}" in c:
+                            todrop.append(c)
+                    self.report.drop(todrop, axis=1, inplace=True)
+#                     del cols, todrop, c
+                    
+                    # Delete dataframe that was already joined and merged into self.report
+                    destroyer=self.subreports[i]
+                    self.subreports[i]=None
+                    del destroyer
+                    
+                    # Force garbage collector
+                    gc.collect()
+                    
+                    buffer = io.StringIO()
+                    self.report.info(verbose=True, buf=buffer)
+                    self.logger.debug("Report memory profile so far:\n{}".format(buffer.getvalue()))
+                    
 
 
 
 
                 # First Stage data conversion - operate over columns
 
-                self.logger.debug("Optimizing data on {} dimensions...".format(len(self.dimensions)))
+                self.logger.debug("Optimizing data types on {} dimensions...".format(len(self.dimensions)))
                 for d in self.dimensions:
                     if 'type' in d:
                         orgname=d['title']
@@ -635,7 +691,9 @@ class GAAPItoDB(object):
                             # Add GA View's time zone just to convert time to UTC right away
                             self.report[d['title']]=self.report[d['title']].apply(lambda x: x.tz_localize(self.gaTimezone).tz_convert(None))
 
-
+                buffer = io.StringIO()
+                self.report.info(verbose=True, buf=buffer)
+                self.logger.debug("Report memory profile after data type optimization:\n{}".format(buffer.getvalue()))
 
                 if self.report.shape[0]>0:
                     # Second Stage data conversion - operate over entire dataframe
@@ -648,6 +706,12 @@ class GAAPItoDB(object):
                             self.report = d['transform'](self.report,d)
 
 
+                    # Sort report by dimension that has synccursor=True
+                    for k in self.dimensions:
+                        if 'synccursor' in k and k['synccursor']==True:
+                            break
+                    self.report.sort_values(by=k['title'], inplace=True)
+
 
                     # Calculate unique IDs for rows
                     self.logger.debug("Generate wanna-be unique IDs for rows...")
@@ -661,8 +725,9 @@ class GAAPItoDB(object):
 
 
                 # Clean the way for more data
+                destroyer = self.report
                 self.report = None
-#                 del self.subreports[:]
+                del destroyer
                 self.subreports.clear()
             
         
@@ -720,14 +785,12 @@ class GAAPItoDB(object):
         
         If `restart` is True, ignore `incremental` and use `start` date.
         """
+        self.effectiveStart = self.start
         
         if self.restart:
             # Ignore lastest data in database and start over
-            self.effectiveStart=self.start
             self.lastSync = None
         else:
-            self.effectiveStart = self.start   # set a default
-
             if self.incremental:
                 # Effective start date is the last date of the `synccursor` column in DB, so find its value.
                 timeColName=None
@@ -751,9 +814,9 @@ class GAAPItoDB(object):
     
     def writeDB(self, rawReport):
         report=rawReport[self.getFinalReportColumns()]
+        timeColName=None
         
-        
-        # For debug purposes, get name of column used as sync parameter
+        # Get name of column used as sync parameter
         for d in self.dimensions:
             if 'synccursor' in d and d['synccursor']:
                 timeColName=d['title']
@@ -772,22 +835,68 @@ class GAAPItoDB(object):
             ifexists='append'
         
         if self.update:
-            # Use pandas.to_sql() to add final report lines to the database table
-            report.reset_index().to_sql(
-                self.targetTable,
-                if_exists=ifexists,
-                index=False,
-                con=self.db
-            )
+            if self.dbWritePartitions and timeColName:
+                partitions=list(report[timeColName].value_counts(bins=self.dbWritePartitions).index.sort_values())
+                for interval in partitions:
+                    if interval.closed_left:
+                        filter = report[timeColName] >= interval.left
+                    else:
+                        filter = report[timeColName] > interval.left
+
+                    if interval.closed_right:
+                        filter &= report[timeColName] <= interval.right
+                    else:
+                        filter &= report[timeColName] < interval.right
+                    
+                    try:
+                        report[filter].reset_index().to_sql(
+                            self.targetTable,
+                            if_exists=ifexists,
+                            index=False,
+                            con=self.db
+                        )
+                    except sqlalchemy.exc.OperationalError as e:
+                        targetFile='bad file for interval {}→{}.csv'.format(interval.left,interval.right)
+                        self.logger.error('Failed to write data partition to DB. Dumping data to CSV for you to check')
+                        report[filter].reset_index().to_csv(targetFile)
+                        self.logger.exception(e)
+                        os._exit(os.EX_DATAERR)
+                    
+                    self.logger.debug('Wrote ({}) {} datapoints of {} columns, ranging from {} to {}'.format(
+                            ifexists,
+                            report[filter].shape[0],
+                            report[filter].shape[1],
+                            report[filter][timeColName].min(),
+                            report[filter][timeColName].max()
+                        )
+                    )
+
+                    
+            else:
+                # Writing in one block, no partitioning (dangerous, may be too much information to your DB)
+                # Use pandas.to_sql() to add final report lines to the database table
+                try:
+                    report.reset_index().to_sql(
+                        self.targetTable,
+                        if_exists=ifexists,
+                        index=False,
+                        con=self.db
+                    )
+                except sqlalchemy.exc.OperationalError as e:
+                    targetFile='bad file for interval {}→{}.csv'.format(report[timeColName].min(),report[timeColName].max())
+                    self.logger.error('Failed to write data to DB. Dumping data to CSV for you to check')
+                    report.reset_index().to_csv(targetFile)
+                    self.logger.exception(e)
+                    os._exit(os.EX_DATAERR)
             
-            self.logger.debug('Wrote ({}) {} datapoints of {} columns, ranging from {} to {}'.format(
-                    ifexists,
-                    report.shape[0],
-                    report.shape[1],
-                    report[timeColName].min(),
-                    report[timeColName].max()
+                self.logger.debug('Wrote ({}) {} datapoints of {} columns, ranging from {} to {}'.format(
+                        ifexists,
+                        report.shape[0],
+                        report.shape[1],
+                        report[timeColName].min(),
+                        report[timeColName].max()
+                    )
                 )
-            )
         else:
             self.logger.warning('Didn’t write ({}) {} datapoints of {} columns, ranging from {} to {}'.format(
                     ifexists,
@@ -824,9 +933,6 @@ class GAAPItoDB(object):
         self.logger.debug('End of DB writing thread.')
 
                 
-            
-    
-
     def sync(self):
         self.connectDB()
         self.effectiveStartDate()   # Sets self.effectiveStart
@@ -835,10 +941,32 @@ class GAAPItoDB(object):
         self.writer = threading.Thread(target=self.databaseWriter)
         self.writer.start() # start the thread
         
-        # Start talking to GA and get report data
-        self.getReportData()
+        try:
+            # Start talking to GA and get report data
+            self.getReportData()
+        except Exception as e:
+            self.logger.exception('GA affairs failed.' + e)
+            os._exit(1)
 
         # Block until there is nothing on the queue
         if not self.dbWriteQueue.empty():
             self.dbWriteQueue.join()
 
+
+    def TransformRegexReplace(self,df,dimension):
+        if 'keeporiginal' in dimension:
+            # Keep original data in a new column with suffix "__org"
+            orgname = dimension['title'] + "__org"
+            df[orgname] = df[dimension['title']]
+
+        self.logger.debug(f"TransformRegexReplace: transforming {dimension['title']}")
+        try:
+            df[dimension['title']]=df[[dimension['title']]].replace(regex=dimension['transformparams'])
+        except Exception as e:
+            self.logger.exception("TransformRegexReplace: unrecoverable error in custom transformation: " + e)
+            os._exit(os.EX_DATAERR)
+
+        self.logger.debug(f"TransformRegexReplace: end of custom transformation.")
+        return df
+    
+
